import copy
from typing import Dict

import mmcv
from easydict import EasyDict as edict
from mmcv.runner.hooks import HOOKS
from pytorch_lightning.callbacks.base import Callback

__all__ = ["LRScheduler"]


class LRScheduler(Callback):

    r"""
    Learning rate scheduler, using mmcv. In mmcv, learning rate
    scheduler is implemented by hook. We need to re-use mmcv
    in lightning, so Callback is used instead.

    Args:
        cfg: learning rate config, in mmcv-fashion.

    Example::

        from pl_extension.callbacks import LRScheduler
        lr_config=dict(
            policy='step',
            warmup='linear',
            warmup_ratio=0.001,
            warmup_iters=1,
            warmup_by_epoch=True,
            step=[21, 26, 28],
            contrib='mmcv',
        )
        lr_callback = LRScheduler(lr_config)
        trainer = Trainer(callbacks=[lr_callback])

    """

    def __init__(self, cfg: Dict):
        lr_config = copy.deepcopy(cfg)
        assert "contrib" in lr_config
        assert lr_config.pop("contrib") == "mmcv"
        # copy-and-paste from mmcv.runner.BaseRunner.register_lr_hook
        assert "policy" in lr_config
        policy_type = lr_config.pop("policy")
        # If the type of policy is all in lower case, e.g., 'cyclic',
        # then its first letter will be capitalized, e.g., to be 'Cyclic'.
        # This is for the convenient usage of Lr updater.
        # Since this is not applicable for `
        # CosineAnnealingLrUpdater`,
        # the string will not be changed if it contains capital letters.
        if policy_type == policy_type.lower():
            policy_type = policy_type.title()
        hook_type = policy_type + "LrUpdaterHook"
        lr_config["type"] = hook_type
        self.hook = mmcv.build_from_cfg(lr_config, HOOKS)
        self.runner = None

    def on_train_start(self, trainer, pl_module):
        """
        before_run in hook.
        """
        self.runner = self.__build_runner_from_trainer(trainer)
        self.hook.before_run(self.runner)

    def on_train_epoch_start(self, trainer, pl_module):
        """
        before_train_epoch in hook.
        """
        self.runner = self.__build_runner_from_trainer(trainer)
        self.runner.epoch = trainer.current_epoch
        self.runner.data_loader = trainer.train_dataloader
        self.hook.before_train_epoch(self.runner)

    def on_train_batch_start(self, trainer, *args, **kwargs):
        """
        before_train_iter in hook.
        """
        self.runner = self.__build_runner_from_trainer(trainer)
        self.runner.epoch = trainer.current_epoch
        self.runner.iter = trainer.global_step
        self.hook.before_train_iter(self.runner)
        # log learning rate
        legacy_metrics = (
            trainer.logger_connector.cached_results.legacy_batch_log_metrics
        )
        legacy_metrics["lr"] = trainer.optimizers[0].param_groups[0]["lr"]

    def __build_runner_from_trainer(self, trainer):
        """
        build runner(mmcv) from trainer(lightning).
        """
        if self.runner:
            return self.runner
        runner = edict()
        # Use first optimizer if multi optimizers provided.
        runner.optimizer = trainer.optimizers[0]
        runner.iter = trainer.global_step
        runner.epoch = trainer.current_epoch
        runner.max_epochs = trainer.max_epochs
        if trainer.max_steps:
            runner.max_iters = trainer.max_steps
        else:
            runner.max_iters = trainer.max_epochs * len(
                trainer.train_dataloader
            )  # noqa
        return runner



import logging

from pytorch_lightning.callbacks.base import Callback

__all__ = ["LinearScalingRule"]

logger = logging.getLogger(__name__)


class LinearScalingRule(Callback):

    r"""
    Linear scaling learning rate rule, following :paper:`ImageNet in 1h`.
    BYW, it also print epoch size.

    Args:
        reference_world_size: reference world size, LinearScalingRule is only
            available when reference_world_size > 0 and 'ddp' is used.

    Example::

        from pl_extension.callbacks import LinearScalingRule
        linear_callback = LinearScalingRule(1)
        trainer = Trainer(callbacks=[linear_callback])

    """

    def __init__(self, reference_world_size: int):
        self.reference_world_size = reference_world_size

    def on_train_start(self, trainer, pl_module):
        self.__show_epoch_size(trainer)
        if self.reference_world_size == 0:
            return
        scale = trainer.world_size // self.reference_world_size
        # TODO: log info about only support ddp mode
        msg = "LinearScalingRule: "
        # batch size
        if hasattr(trainer.datamodule, "batch_size"):
            # batch_size is not multiplied by scale, because
            # trainer.datamodule.batch_size is actually batch_size_per_gpu.
            # When in ddp, total_batch_size = batch_size_per_gpu * scale,
            # which is done by ddp, so we skip here.
            msg, _ = self.__update(
                trainer.datamodule.batch_size,
                trainer.datamodule.batch_size * scale,
                msg,
                "batch_size",
            )
            total_batch_size = trainer.datamodule.batch_size * scale
        else:
            total_batch_size = scale
        # learning rate
        lr_scale = total_batch_size
        msg, lr = self.__update(
            trainer.optimizers[0].param_groups[0]["lr"],
            trainer.optimizers[0].param_groups[0]["lr"] * lr_scale,
            msg,
            "learning_rate",
        )
        trainer.optimizers[0].param_groups[0]["lr"] = lr
        # max steps
        if trainer.max_steps:
            msg, max_steps = self.__update(
                trainer.max_steps, trainer.max_steps // scale, msg, "max steps"
            )
            trainer.max_steps = max_steps
        # min steps
        if trainer.min_steps:
            msg, min_steps = self.__update(
                trainer.min_steps, trainer.min_steps // scale, msg, "min steps"
            )
            trainer.min_steps = min_steps
        # TODO: warmup steps
        logger.info(msg[:-2])  # ignore last ", "

    def __update(self, old_value, new_value, msg, name):
        msg += f"{name}: {old_value} => {new_value}, "
        return msg, new_value

    def __show_epoch_size(self, trainer):
        epoch_size = trainer.num_training_batches
        logger.info(f"epoch size: {epoch_size}")



import logging

from pl_extension.loggers import logging_logger


def test_logginglogger(tmpdir):
    logger = logging_logger.LoggingLogger(
        tmpdir, prefix="ple", level=logging.INFO
    )
    logger.info("hello, pl-extension!")



import logging

from pytorch_lightning.callbacks.base import Callback

__all__ = ["Evaluation"]

logger = logging.getLogger(__name__)


class Evaluation(Callback):

    r"""
    Evaluation callback. Do evaluation after testing.
    """

    def on_validation_start(self, trainer, pl_module):
        if trainer.running_sanity_check or not pl_module.do_validation:
            return
        logger.info("Run validation ...")

    def on_validation_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
    ):
        if len(trainer.num_val_batches) > 1:
            raise NotImplementedError
        if trainer.running_sanity_check or not pl_module.do_validation:
            return
        num_batch = trainer.num_val_batches[0]
        if batch_idx % pl_module.log_every_n_steps_in_valid == 0:
            logger.info(f"{batch_idx} / {num_batch}")

    def on_validation_end(self, trainer, pl_module):
        """
        Do evaluation when valid finished.
        """
        if trainer.running_sanity_check or not pl_module.do_validation:
            return
        if not pl_module.evaluation_when_valid:
            return
        if hasattr(pl_module, "report_metric"):
            pl_module.report_metric()

    def on_test_start(self, trainer, pl_module):
        logger.info("Run test ...")

    def on_test_batch_end(
        self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx
    ):
        if len(trainer.num_test_batches) > 1:
            raise NotImplementedError
        num_batch = trainer.num_test_batches[0]
        if batch_idx % pl_module.log_every_n_steps_in_test == 0:
            logger.info(f"{batch_idx} / {num_batch}")

    def on_test_end(self, trainer, pl_module):
        """
        Do evaluation when test finished.
        """
        if hasattr(pl_module, "report_metric"):
            pl_module.report_metric()



import os

from pytorch_lightning.loggers import MLFlowLogger as _MLFlowLogger

__all__ = ["MLFlowLogger"]


class MLFlowLogger(_MLFlowLogger):

    """Custom run_id for MLFlowLogger, required by AIDI."""

    @property
    def run_id(self):
        _ = self.experiment
        _run_id = os.environ.get("MLFLOW_RUN_ID", self._run_id)
        return _run_id



import time

__all__ = ["time_string"]


def time_string(short=False) -> str:
    """
    Return time as string, used as random string
    """
    year = time.localtime().tm_year
    mon = time.localtime().tm_mon
    day = time.localtime().tm_mday
    hour = time.localtime().tm_hour
    min = time.localtime().tm_min
    sec = time.localtime().tm_sec
    if short:
        return "%02d%02d%02d" % (hour, min, sec)
    else:
        return "%04d%02d%02d_%02d%02d%02d" % (year, mon, day, hour, min, sec)



import argparse
import logging
from typing import Any, Dict

__all__ = ["add_options", "apply_options", "update_config"]

logger = logging.getLogger(__name__)


def add_options(parser: argparse.ArgumentParser) -> None:
    """
    Add options to parser.
    """
    parser.add_argument(
        "opts",
        help="modify config options using the command-line",
        default=None,
        nargs=argparse.REMAINDER,
    )


def apply_options(config: Dict[str, Any], args: argparse.Namespace) -> None:
    """
    Update config using opts.
    """
    try:
        from yacs.config import CfgNode as CN
    except ImportError as e:
        logger.error(
            "`apply_options` requiring yacs(version==0.1.8), "
            "but not installed."
        )
        raise e
    opts = args.opts
    config_CN = CN(config)
    if opts is not None and len(opts) > 0:
        config_CN.merge_from_list(opts)
    config.update(config_CN)


def update_config(config: Dict[str, Any], vars_dict: Dict[str, Any]) -> None:
    """Update str in config, with vars_dict.

    When you want to set some lazy-compute value in config, `update_config` is
    very useful.

    Args:
        config: input config dict.
        vars_dict: key-value pair.

    Examples:

        config = dict(
            a="python",
            b="FILENAME",
            c=["python", "FILENAME", "python FILENAME"],
        )
        vars_dict = {"FILENAME": "rust"}
        update_config(config, vars_dict)
        assert config["b"] == "rust"
        assert config["c"][1] == "rust"
        assert config["c"][2] == "python rust"

    """

    def _find_and_replace(_str):
        for _k, _v in vars_dict.items():
            _str = _str.replace(_k, _v)
        return _str

    if isinstance(config, dict):
        for k in config:
            if isinstance(config[k], str):
                config[k] = _find_and_replace(config[k])
            elif isinstance(config[k], dict):
                update_config(config[k], vars_dict)
            elif isinstance(config[k], list):
                update_config(config[k], vars_dict)
    elif isinstance(config, list):
        for i, _ in enumerate(config):
            if isinstance(config[i], str):
                config[i] = _find_and_replace(config[i])
            elif isinstance(config[i], dict):
                update_config(config[i], vars_dict)
            elif isinstance(config[i], list):
                update_config(config[i], vars_dict)
    elif isinstance(config, (int, float, bool)):
        pass
    else:
        raise ValueError(f"Unknown type: {type(config)}")



import os
from distutils.version import LooseVersion

PL_EXTENSION_CACHE = os.getenv(
    "PL_EXTENSION_CACHE", os.path.expanduser("~/.cache/pl_extension/")
)

HDFS_CACHE = PL_EXTENSION_CACHE


def check_version(name, version, strict=False):
    if strict:
        status = LooseVersion(name.__version__) == LooseVersion(version)
        msg = (
            f"{name.__name__}=={version} is required, "
            f"but found {name.__version__}."
        )
    else:
        status = LooseVersion(name.__version__) >= LooseVersion(version)
        msg = (
            f"{name.__name__}>={version} is required, "
            f"but found {name.__version__}."
        )
    if not status:
        raise ImportError(msg)



from pl_extension.utilities import rand
from pl_extension.utilities.config import update_config


def test_time_string():
    rand_str = rand.time_string()
    assert rand_str


def test_update_config():
    config = dict(
        a="python",
        b="FILENAME",
        c=["python", "FILENAME", "python FILENAME"],
    )
    vars_dict = {"FILENAME": "rust"}
    update_config(config, vars_dict)
    assert config["b"] == "rust"
    assert config["c"][1] == "rust"
    assert config["c"][2] == "python rust"
